package net.seesaa.nekokiss.algorithm.make_maze.extend_wall;

import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.Dimension;
import java.awt.EventQueue;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.Font;
import java.awt.Graphics;
import java.awt.GridBagConstraints;
import java.awt.GridBagLayout;
import java.awt.image.BufferedImage;
import java.awt.Point;

import java.util.LinkedList;
import java.util.Queue;

import javax.swing.border.EmptyBorder;
import javax.swing.border.EtchedBorder;
import javax.swing.border.TitledBorder;
import javax.swing.event.ChangeEvent;
import javax.swing.event.ChangeListener;
import javax.swing.JButton;
import javax.swing.JCheckBox;
import javax.swing.JComboBox;
import javax.swing.JComponent;
import javax.swing.JFrame;
import javax.swing.JLabel;
import javax.swing.JPanel;
import javax.swing.JScrollPane;
import javax.swing.JSpinner;
import javax.swing.SpinnerNumberModel;
import javax.swing.SwingConstants;
import javax.swing.Timer;

final class VisualizedFrame extends JFrame implements ActionListener, Runnable {

	private static final class ILLabel extends JLabel implements Runnable {

		private String text;

		ILLabel() { }

		@Override
		public final void run() {
			this.setText(this.text);
		}

		final void setTextIL(String text) {
			this.text = text;
			EventQueue.invokeLater(this);
		}

	}

	private final class MazePainter extends JComponent implements Runnable {

		private Dimension size;

		MazePainter() { }

		@Override
		protected final void paintComponent(Graphics g) {
			VisualizedFrame.this.paintMaze(g);
		}

		@Override
		public final void run() {
			this.setPreferredSize(this.size);
			this.revalidate();
		}

		final void setSizeIL(int width, int height) {
			this.size = new Dimension(width, height);
			EventQueue.invokeLater(this);
		}

	}

	private static final class SpinnerOddNumberModel extends SpinnerNumberModel
			implements ChangeListener {

		SpinnerOddNumberModel(int value, int minimum, int maximum,
				int stepSize) {
			super(value, minimum, maximum, stepSize);
			this.addChangeListener(this);
		}

		@Override
		public final void stateChanged(ChangeEvent e) {
			int v = (int)this.getValue();
			if ((v & 1) == 0) {
				this.setValue(--v);
			}
		}

	}

	private final Queue<Point> wallList = new LinkedList<>();
	private final Timer timer;
	private final ILLabel sizeLabel;
	private final MazePainter mazePainter;
	private final JCheckBox animationCheckBox;
	private final SpinnerNumberModel delayModel;
	private final JCheckBox wallTypeBox;
	private final SpinnerNumberModel blockMazeWidthModel;
	private final SpinnerNumberModel blockMazeHeightModel;
	private final JComboBox blockSizeBox;
	private final SpinnerNumberModel thinMazeWidthModel;
	private final SpinnerNumberModel thinMazeHeightModel;
	private final JCheckBox rndCheckBox;
	private final SpinnerNumberModel xModel;
	private final SpinnerNumberModel yModel;
	private final SpinnerNumberModel zModel;
	private final SpinnerNumberModel wModel;
	private int width;
	private int height;
	private Maze maze;
	private boolean busy;
	private BufferedImage img;
	private boolean blockWall;
	private int blockSize;
	private int x;
	private int y;
	private int z;
	private int w;

	private VisualizedFrame() {
		super("壁伸ばし法迷路生成視覚化 1.01");
		this.setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);
		
		Font m12Font = new Font(Font.MONOSPACED, Font.PLAIN, 12);
		Font m15Font = new Font(Font.MONOSPACED, Font.PLAIN, 15);
		Font m18Font = new Font(Font.MONOSPACED, Font.PLAIN, 18);

		this.timer = new Timer(25, this);

		JComponent c = (JComponent)this.getContentPane();
		c.setBorder(new EmptyBorder(4, 4, 4, 4));

		JComponent lPnl = new JPanel(new BorderLayout());
		c.add(lPnl);

		this.sizeLabel = new ILLabel();
		this.sizeLabel.setHorizontalAlignment(SwingConstants.RIGHT);
		this.sizeLabel.setBorder(new EtchedBorder());
		this.sizeLabel.setFont(m18Font);
		lPnl.add(this.sizeLabel, BorderLayout.NORTH);
		this.updateSizeLabel();

		this.mazePainter = new MazePainter();

		JScrollPane sPane = new JScrollPane(this.mazePainter);
		sPane.setPreferredSize(new Dimension(420, 360));
		lPnl.add(sPane);

		GridBagLayout gbl = new GridBagLayout();

		JComponent rPnl = new JPanel(gbl);
		rPnl.setBorder(new EtchedBorder());
		c.add(rPnl, BorderLayout.EAST);

		GridBagConstraints gbc = new GridBagConstraints();

		JButton btn = new JButton("生成");
		btn.addActionListener(new ActionListener() {
			@Override
            public final void actionPerformed(ActionEvent e) {
				VisualizedFrame.this.generate();
			}
		});
		btn.setFont(m18Font);
		gbc.gridx = 0;
		gbc.gridy = 0;
		gbc.gridwidth = 2;
		gbl.setConstraints(btn, gbc);
		rPnl.add(btn);

		this.animationCheckBox = new JCheckBox("アニメーション", true);
		this.animationCheckBox.setFont(m15Font);
		++gbc.gridy;
		gbl.setConstraints(this.animationCheckBox, gbc);
		rPnl.add(this.animationCheckBox);

		JLabel lbl = new JLabel("描画速度");
		lbl.setFont(m18Font);
		++gbc.gridy;
		--gbc.gridwidth;
		gbl.setConstraints(lbl, gbc);
		rPnl.add(lbl);

		this.delayModel = new SpinnerNumberModel(1, 1, 250, 1);

		JComponent spn = new JSpinner(this.delayModel);
		spn.setFont(m15Font);
		++gbc.gridx;
		gbl.setConstraints(spn, gbc);
		rPnl.add(spn);

		btn = new JButton("休止/再開");
		btn.addActionListener(new ActionListener() {
			@Override
            public final void actionPerformed(ActionEvent e) {
				VisualizedFrame.this.pause();
			}
		});
		btn.setFont(m18Font);
		--gbc.gridx;
		++gbc.gridy;
		++gbc.gridwidth;
		gbl.setConstraints(btn, gbc);
		rPnl.add(btn);

		btn = new JButton("中止");
		btn.addActionListener(new ActionListener() {
			@Override
            public final void actionPerformed(ActionEvent e) {
				VisualizedFrame.this.endVisualize();
			}
		});
		btn.setFont(m18Font);
		++gbc.gridy;
		gbl.setConstraints(btn, gbc);
		rPnl.add(btn);

		this.wallTypeBox = new JCheckBox("薄い壁");
		this.wallTypeBox.setFont(m18Font);
		++gbc.gridy;
		gbl.setConstraints(this.wallTypeBox, gbc);
		rPnl.add(this.wallTypeBox);

		GridBagLayout gbl2 = new GridBagLayout();

		TitledBorder tb = new TitledBorder(new EtchedBorder(EtchedBorder
				.LOWERED), "ブロック壁");
		tb.setTitleFont(m12Font);

		JComponent pnl1 = new JPanel(gbl2);
		pnl1.setBorder(tb);
		++gbc.gridy;
		gbc.fill = GridBagConstraints.HORIZONTAL;
		gbl.setConstraints(pnl1, gbc);
		rPnl.add(pnl1);

		GridBagLayout gbl3 = new GridBagLayout();

		tb = new TitledBorder(new EtchedBorder(EtchedBorder
				.LOWERED), "薄い壁");
		tb.setTitleFont(m12Font);

		JComponent pnl2 = new JPanel(gbl3);
		pnl2.setBorder(tb);
		++gbc.gridy;
		gbl.setConstraints(pnl2, gbc);
		rPnl.add(pnl2);

		GridBagLayout gbl4 = new GridBagLayout();

		tb = new TitledBorder(new EtchedBorder(EtchedBorder
				.LOWERED), "xor128");
		tb.setTitleFont(m12Font);

		JComponent pnl3 = new JPanel(gbl4);
		pnl3.setBorder(tb);
		++gbc.gridy;
		gbc.anchor = GridBagConstraints.NORTH;
		gbc.weighty = 1.0d;
		gbl.setConstraints(pnl3, gbc);
		rPnl.add(pnl3);

		lbl = new JLabel("Width");
		lbl.setFont(m18Font);
		gbc.gridy = 0;
		--gbc.gridwidth;
		gbc.fill = GridBagConstraints.NONE;
		gbc.anchor = GridBagConstraints.WEST;
		gbc.weightx = 1.0d;
		gbc.weighty = 0;
		gbl2.setConstraints(lbl, gbc);
		pnl1.add(lbl);

		this.blockMazeWidthModel = new SpinnerOddNumberModel(25, 5, 255, 2);

		spn = new JSpinner(this.blockMazeWidthModel);
		spn.setFont(m15Font);
		++gbc.gridx;
		gbc.anchor = GridBagConstraints.EAST;
		gbc.weightx = 0;
		gbl2.setConstraints(spn, gbc);
		pnl1.add(spn);

		lbl = new JLabel("Height");
		lbl.setFont(m18Font);
		--gbc.gridx;
		++gbc.gridy;
		gbc.anchor = GridBagConstraints.WEST;
		gbc.weightx = 1.0d;
		gbl2.setConstraints(lbl, gbc);
		pnl1.add(lbl);

		this.blockMazeHeightModel = new SpinnerOddNumberModel(25, 5, 255, 2);

		spn = new JSpinner(this.blockMazeHeightModel);
		spn.setFont(m15Font);
		++gbc.gridx;
		gbc.anchor = GridBagConstraints.EAST;
		gbc.weightx = 0;
		gbl2.setConstraints(spn, gbc);
		pnl1.add(spn);

		lbl = new JLabel("Block Size");
		lbl.setFont(m18Font);
		--gbc.gridx;
		++gbc.gridy;
		gbc.anchor = GridBagConstraints.WEST;
		gbl2.setConstraints(lbl, gbc);
		pnl1.add(lbl);

		this.blockSizeBox = new JComboBox(new String[] {
			"1", "2", "4", "8"
		});
		this.blockSizeBox.setFont(m15Font);
		this.blockSizeBox.setSelectedIndex(2);
		++gbc.gridx;
		gbc.anchor = GridBagConstraints.EAST;
		gbl2.setConstraints(this.blockSizeBox, gbc);
		pnl1.add(this.blockSizeBox);

		lbl = new JLabel("Width");
		lbl.setFont(m18Font);
		--gbc.gridx;
		gbc.gridy = 0;
		gbc.anchor = GridBagConstraints.WEST;
		gbc.weightx = 1.0d;
		gbl3.setConstraints(lbl, gbc);
		pnl2.add(lbl);

		this.thinMazeWidthModel = new SpinnerNumberModel(12, 2, 127, 1);

		spn = new JSpinner(this.thinMazeWidthModel);
		spn.setFont(m15Font);
		++gbc.gridx;
		gbc.weightx = 0;
		gbc.anchor = GridBagConstraints.EAST;
		gbl3.setConstraints(spn, gbc);
		pnl2.add(spn);

		lbl = new JLabel("Height");
		lbl.setFont(m18Font);
		--gbc.gridx;
		++gbc.gridy;
		gbc.weightx = 1.0d;
		gbc.anchor = GridBagConstraints.WEST;
		gbl3.setConstraints(lbl, gbc);
		pnl2.add(lbl);

		this.thinMazeHeightModel = new SpinnerNumberModel(12, 2, 127, 1);

		spn = new JSpinner(this.thinMazeHeightModel);
		spn.setFont(m15Font);
		++gbc.gridx;
		gbc.weightx = 0;
		gbc.anchor = GridBagConstraints.EAST;
		gbl3.setConstraints(spn, gbc);
		pnl2.add(spn);

		this.rndCheckBox = new JCheckBox("Seed を指定する");
		this.rndCheckBox.setFont(m18Font);
		--gbc.gridx;
		gbc.gridy = 0;
		++gbc.gridwidth;
		gbc.anchor = GridBagConstraints.CENTER;
		gbl4.setConstraints(this.rndCheckBox, gbc);
		pnl3.add(this.rndCheckBox);

		lbl = new JLabel("x");
		lbl.setFont(m18Font);
		++gbc.gridy;
		--gbc.gridwidth;
		gbc.anchor = GridBagConstraints.WEST;
		gbc.weightx = 1.0d;
		gbl4.setConstraints(lbl, gbc);
		pnl3.add(lbl);

		this.xModel = new SpinnerNumberModel(0, 0, 0xffffffffL, 1);

		spn = new JSpinner(this.xModel);
		spn.setFont(m15Font);
		++gbc.gridx;
		gbc.weightx = 0;
		gbc.anchor = GridBagConstraints.EAST;
		gbl4.setConstraints(spn, gbc);
		pnl3.add(spn);

		lbl = new JLabel("y");
		lbl.setFont(m18Font);
		--gbc.gridx;
		++gbc.gridy;
		gbc.anchor = GridBagConstraints.WEST;
		gbc.weightx = 1.0d;
		gbl4.setConstraints(lbl, gbc);
		pnl3.add(lbl);

		this.yModel = new SpinnerNumberModel(0, 0, 0xffffffffL, 1);

		spn = new JSpinner(this.yModel);
		spn.setFont(m15Font);
		++gbc.gridx;
		gbc.weightx = 0;
		gbc.anchor = GridBagConstraints.EAST;
		gbl4.setConstraints(spn, gbc);
		pnl3.add(spn);

		lbl = new JLabel("z");
		lbl.setFont(m18Font);
		--gbc.gridx;
		++gbc.gridy;
		gbc.anchor = GridBagConstraints.WEST;
		gbc.weightx = 1.0d;
		gbl4.setConstraints(lbl, gbc);
		pnl3.add(lbl);

		this.zModel = new SpinnerNumberModel(0, 0, 0xffffffffL, 1);

		spn = new JSpinner(this.zModel);
		spn.setFont(m15Font);
		++gbc.gridx;
		gbc.weightx = 0;
		gbc.anchor = GridBagConstraints.EAST;
		gbl4.setConstraints(spn, gbc);
		pnl3.add(spn);

		lbl = new JLabel("w");
		lbl.setFont(m18Font);
		--gbc.gridx;
		++gbc.gridy;
		gbc.anchor = GridBagConstraints.WEST;
		gbc.weightx = 1.0d;
		gbl4.setConstraints(lbl, gbc);
		pnl3.add(lbl);

		this.wModel = new SpinnerNumberModel(0, 0, 0xffffffffL, 1);

		spn = new JSpinner(this.wModel);
		spn.setFont(m15Font);
		++gbc.gridx;
		gbc.weightx = 0;
		gbc.anchor = GridBagConstraints.EAST;
		gbl4.setConstraints(spn, gbc);
		pnl3.add(spn);

		this.pack();

	}

	public static final void main(String[] args) {
		EventQueue.invokeLater(new Runnable() {
			@Override
			public final void run() {
				try {
					JFrame frame = new VisualizedFrame();
					frame.setVisible(true);
					frame.setLocationRelativeTo(null);
				} catch (Exception e) {
					e.printStackTrace();
				}
			}
		});
	}
	
	@Override
	public final void actionPerformed(ActionEvent e) {
		if (this.wallList.isEmpty()) {
			this.endVisualize();
			return;
		}
		if (this.blockWall) {
			if (this.blockSize > 1) {
				Point p1 = this.wallList.remove();
				Point p2 = this.wallList.remove();
				Point p;
				int w = this.blockSize;
				int h = this.blockSize;
				if (p1.x > p2.x) {
					// ←
					p = p2;
					w <<= 1;
				} else if (p1.x < p2.x) {
					// →
					p = p1;
					w <<= 1;
				} else if (p1.y > p2.y) {
					// ↑
					p = p2;
					h <<= 1;
				} else {
					// ↓
					p = p1;
					h <<= 1;
				}
				Graphics g = this.img.getGraphics();
				g.setColor(Color.BLACK);
				g.fillRect(p.x * this.blockSize, p.y * this.blockSize, w, h);
				g.dispose();
			} else {
				Point p;
				for (int i = 0; i < 2; ++i) {
					p = this.wallList.remove();
					this.img.setRGB(p.x, p.y, 0);
				}
			}
		} else {
			Point p1 = this.wallList.remove();
			Point p2 = this.wallList.remove();
			Graphics g = this.img.getGraphics();
			g.setColor(Color.BLACK);
			if (p1.x > p2.x) {
				// ←
				++p1.x;
			} else if (p1.x < p2.x) {
				// →
				--p1.x;
			} else if (p1.y > p2.y) {
				// ↑
				++p1.y;
			} else {
				// ↓
				--p1.y;
			}
			g.drawLine(p1.x << 2, p1.y << 2, p2.x << 2, p2.y << 2);
			g.dispose();
		}
		this.mazePainter.repaint();
	}

	@Override
	public final void run() {
		this.xModel.setValue((long)(this.x & 0xffffffffL));
		this.yModel.setValue((long)(this.y & 0xffffffffL));
		this.zModel.setValue((long)(this.z & 0xffffffffL));
		this.wModel.setValue((long)(this.w & 0xffffffffL));
	}

	private final void paintMaze(Graphics g) {
		if (this.img != null) {
			g.drawImage(this.img, 0, 0, this.width, this.height, null);
		}
	}

	private final void formatMaze(int width, int height) {
		this.maze = new Maze(width, height) {
			@Override
			protected final void setWalls(Point p1, Point p2) {
				super.setWalls(p1, p2);
				VisualizedFrame.this.wallList.add(p1);
				VisualizedFrame.this.wallList.add(p2);
			}
		};
	}

	private final void updateSizeLabel() {
		this.sizeLabel.setTextIL(String.format("Width %3d, Height %3d",
				this.width, this.height));
	}

	private final void updateSeeds() {
		this.x = Maze.RND.getX();
		this.y = Maze.RND.getY();
		this.z = Maze.RND.getZ();
		this.w = Maze.RND.getW();
		EventQueue.invokeLater(this);
	}

	private final void pause() {
		if (this.busy) {
			if (this.timer.isRunning()) {
				this.timer.stop();
			} else {
				this.timer.setDelay(this.delayModel.getNumber().intValue());
				this.timer.restart();
			}
		}
	}

	private final void generate() {
		if (this.busy) {
			return;
		}
		this.busy = true;
		if (!this.wallList.isEmpty()) {
			this.wallList.clear();
		}		
		switch (this.blockSizeBox.getSelectedIndex()) {
		case 0:	// 1
			this.blockSize = 1;
			break;
		case 1:	// 2
			this.blockSize = 2;
			break;
		default:
		case 2:	// 4
			this.blockSize = 4;
			break;
		case 3:	// 8
			this.blockSize = 8;
			break;
		}
		this.blockWall = !this.wallTypeBox.isSelected();
		if (this.blockWall) {
			this.width = this.blockMazeWidthModel.getNumber().intValue();
			this.height = this.blockMazeHeightModel.getNumber().intValue();
			this.thinMazeWidthModel.setValue(this.width >> 1);
			this.thinMazeHeightModel.setValue(this.height >> 1);
		} else {
			this.width = this.thinMazeWidthModel.getNumber().intValue();
			this.height = this.thinMazeHeightModel.getNumber().intValue();
			this.blockMazeWidthModel.setValue((this.width << 1) + 1);
			this.blockMazeHeightModel.setValue((this.height << 1) + 1);
		}
		this.updateSizeLabel();
		if (!this.blockWall) {
			this.width = (this.width << 1) + 1;
			this.height = (this.height << 1) + 1;
		}
		this.formatMaze(this.width, this.height);
		if (this.blockWall) {
			if (this.blockSize > 1) {
				this.width *= this.blockSize;
				this.height *= this.blockSize;
			}
			this.mazePainter.setSizeIL(this.width, this.height);
		} else {
			this.width = (this.width >> 1 << 3) + 1;
			this.height = (this.height >> 1 << 3) + 1;
			this.mazePainter.setSizeIL(this.width, this.height);
		}
		if (this.rndCheckBox.isSelected()) {
			this.x = this.xModel.getNumber().intValue();
			this.y = this.yModel.getNumber().intValue();
			this.z = this.zModel.getNumber().intValue();
			this.w = this.wModel.getNumber().intValue();
			if ((x | y | z | w) == 0) {
				this.updateSeeds();
			} else {
				Maze.RND.setSeeds(x, y, z, w);
			}
		} else {
			this.updateSeeds();
		}
		this.maze.generate();
		this.beginVisualize();
	}

	private final void beginVisualize() {
		this.busy = true;
		this.img = new BufferedImage(this.width, this.height,
				BufferedImage.TYPE_INT_RGB);
		Graphics g = this.img.getGraphics();
		g.setColor(Color.WHITE);
		if (this.blockWall && this.blockSize > 1) {
			int b2 = (this.blockSize << 1);
			g.fillRect(this.blockSize, this.blockSize, this.width - b2,
					this.height - b2);
		} else {
			g.fillRect(1, 1, this.width - 2, this.height - 2);
		}
		if (!this.animationCheckBox.isSelected()) {
			if (this.blockWall) {
				if (this.blockSize > 1) {
					g.setColor(Color.BLACK);
					for (Point p: this.wallList) {
						g.fillRect(p.x * this.blockSize, p.y * this.blockSize,
								this.blockSize, this.blockSize);
					}
					g.dispose();
				} else {
					g.dispose();
					for (Point p: this.wallList) {
						this.img.setRGB(p.x, p.y, 0);
					}
				}
			} else {
				g.setColor(Color.BLACK);
				Point p1, p2;
				for (int i = 0, max = this.wallList.size() >> 1; i < max; ++i) {
					p1 = this.wallList.remove();
					p2 = this.wallList.remove();
					if (p1.x > p2.x) {
						// ←
						++p1.x;
					} else if (p1.x < p2.x) {
						// →
						--p1.x;
					} else if (p1.y > p2.y) {
						// ↑
						++p1.y;
					} else {
						// ↓
						--p1.y;
					}
					g.drawLine(p1.x << 2, p1.y << 2, p2.x << 2, p2.y << 2);
				}
				g.dispose();
			}
			this.mazePainter.repaint();
			this.busy = false;
			return;
		} else {
			g.dispose();
			this.timer.setDelay(this.delayModel.getNumber().intValue());
		}
		this.mazePainter.repaint();
		this.timer.start();
	}

	private final void endVisualize() {
		this.timer.stop();
		this.busy = false;
	}

}

